package edu.stanford.db.rdf.uml;

import org.w3c.rdf.model.*;
import org.w3c.rdf.implementation.model.NodeFactoryImpl;
import org.w3c.rdf.util.*;
import java.util.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import org.w3c.rdf.vocabulary.rdf_schema_19990303.RDFS;
import edu.stanford.db.rdf.uml.vocabulary.uml_core_20000422.UMLCore;
import edu.stanford.db.rdf.uml.vocabulary.uml_datatypes_20000507.DataTypes;
import edu.stanford.db.rdf.schema.InvalidModelException;
import edu.stanford.db.rdf.schema.RDFSchemaModel;

/**
 * A UML model supports inverse properties and generalization and validates UML schemas.
 * For example, the following sets of objects/subjects are equivalent:
 * <p><tt>find(X, UMLCore.Feature_owner, null)</tt> or
 * <p><tt>find(null, UMLCore.Classifier_feature, X)</tt>
 * <p>To determine whether X is a class, one can use:
 * <p><tt>find(X, RDF.type, UMLCore.Class)</tt>
 * Validation supports typing and cardinality constraints. Currently, we do not check
 * for ordered relationships.
 */

public class UMLModel implements VirtualModel {

  NodeFactory nodeFactory;
  Model instances, schema;
  Hashtable inverse; // may be shared between different UMLModels

  private UMLModel(NodeFactory f, Hashtable inverse) {
    this.inverse = inverse;
    this.nodeFactory = f;
  }

  public UMLModel(NodeFactory f) {
    this(f, new Hashtable());
  }

  public UMLModel() {
    this(new NodeFactoryImpl());
  }

  public UMLModel(Model instances, Model schema) {
    this();
    this.instances = instances;
    this.schema = schema;
  }

  public UMLModel(NodeFactory f, Model instances, Model schema) {
    this(f);
    this.instances = instances;
    this.schema = schema;
  }

  private UMLModel(NodeFactory f, Model instances, Model schema, Hashtable inverse) {
    this(f, inverse);
    this.instances = instances;
    this.schema = schema;
  }


  /** returns "A subClassOf B" instead of "R parent A & R child B"
   * schema should be UMLModel (must not if generalization and specialization are not used)
   */
  public static Model getRDFSClassHierarchy(Model m) throws ModelException {

    Model res = m.create();
    NodeFactory nodeFactory = m.getNodeFactory();
    // find all generalizations
    Model gen = m.find(null, RDF.type, UMLCore.Generalization);
    //    System.err.println("Generalizations in the schema: " + gen.size());
    for(Enumeration en = gen.elements(); en.hasMoreElements();) {
      Resource g = ((Statement)en.nextElement()).subject();
      // find parent and child using UMLModel (i.e. inverse relationships would work, too)
      Resource parent = RDFUtil.getObjectResource(m, g, UMLCore.Generalization_parent);
      Resource child = RDFUtil.getObjectResource(m, g, UMLCore.Generalization_child);
      if(parent == null)
        throw new InvalidModelException("[UMLModel] no parent in Generalization " + g);
      else if(child == null)
        throw new InvalidModelException("[UMLModel] no child in Generalization " + g);
      else {
	//	System.out.println("SUBCLASSOF: " + child + " --> " + parent);
        res.add(nodeFactory.createStatement(child, RDFS.subClassOf, parent));
      }
    }
    return res;
  }

  // precondition: inverse is an empty Hashtable

  private Hashtable getInverse() throws ModelException {

    synchronized(inverse) {
      if(inverse.isEmpty())
	inverse = computeInverse(schema);
      return inverse;
    }
  }

  /** Computes hashtable containing the names of the inverse relationships */

  public static Hashtable computeInverse(Model schema) throws ModelException {

    Hashtable inverse = new Hashtable();

    Model ass = schema.find(null, RDF.type, UMLCore.Association);
    //                System.err.println("Associations in the schema " + this + ": " + ass.size());
    //                new Exception().printStackTrace(System.err);
    for(Enumeration en = ass.elements(); en.hasMoreElements();) {
      Resource a = ((Statement)en.nextElement()).subject();
      Model seq = schema.find(a, UMLCore.Association_connection, null);
      if(seq.size() != 2)
	throw new InvalidModelException("[UMLModel] cardinality violation: " + seq.size() + " instead of 2 objects for property " + UMLCore.Association_connection + " found for " + a);
      // FIXME: order is not really needed here... Otherwise could use org.w3c.rdf.util.OrderUtil
      Enumeration seqEn = seq.elements();
      Resource first = (Resource)((Statement)seqEn.nextElement()).object();
      Resource second = (Resource)((Statement)seqEn.nextElement()).object();

      //      System.out.println("INVERSE: " + first + " <--> " + second);
      inverse.put(first, second);
      inverse.put(second, first);

      /*
      if(seq != null) {
	Resource first = RDFUtil.getObjectResource(schema, seq, schema.getNodeFactory().createOrdinal(1));
	Resource second = RDFUtil.getObjectResource(schema, seq, schema.getNodeFactory().createOrdinal(2));
	if(first == null)
	  throw new InvalidModelException("[UMLModel] no first element from " + seq);
	else if(second == null)
	  throw new InvalidModelException("[UMLModel] no second element from " + seq);
	else {
	  inverse.put(first, second);
	  inverse.put(second, first);
	}
      } else
	throw new InvalidModelException("[UMLModel] no Association.connection from " + a);
      */
    }
    //                System.err.println("Total inverse elements: " + inverse.size() / 2);
    return inverse;
  }

  public String getLabel() throws ModelException {

    return instances.getLabel();
  }

  public String getURI() throws ModelException {

    return instances.getURI();
  }

  /**
   * @return model contains the fact basis of this model
   */
  public Model getGroundModel() throws ModelException {
    return instances;
  }

  /**
   * Set a base URI for the model.
   * Affects creating of new resources and serialization syntax.
   * Inherited method getURI returns the URI set in this method
   */
  public void setSourceURI(String uri) throws ModelException {
    instances.setSourceURI(uri);
  }

  /**
   * Returns current base URI setting.
   */
  public String getSourceURI() throws ModelException {
    return instances.getSourceURI();
  }

  // Model access

  /**
   * Number of triples in the model
   *
   * @return  number of triples, -1 if unknown
   * 
   * @seeAlso org.w3c.rdf.model.VirtualModel
   */
  public int size() throws ModelException {
    return -1;
  }

  /**
   * true if the model contains no triples
   */
  public boolean isEmpty() throws ModelException {
    return instances.isEmpty();
  }

  /**
   * Enumerate triples
   */
  public Enumeration elements() throws ModelException {

    return find(null, null, null).elements();
  }

  /**
   * Tests if the model contains the given triple.
   *
   * @return  <code>true</code> if the triple belongs to the model;
   *          <code>false</code> otherwise.
   */
  public boolean contains(Statement t) throws ModelException {
    // FIXME: efficiency?
    return !find(t.subject(), t.predicate(), t.object()).isEmpty();
  }

  // Model manipulation: add, remove, find

  /**
   * Adds a new triple to the model.
   */
  public void add(Statement t) throws ModelException {
    instances.add(t);
  }

  /**
   * Removes the triple from the model.
   */
  public void remove(Statement t) throws ModelException {
    instances.remove(t);
  }

  public boolean isMutable() throws ModelException {

    return instances.isMutable();
  }

  /**
   * General method to search for triples.
   */
  public Model find( Resource subject, Resource predicate, RDFNode object ) throws ModelException {

    getInverse();

    // we need it anyway
    Model res = instances.find(subject, predicate, object);

    Model derived = res.create();
    if(predicate != null) {
      collectInverse(derived, subject, predicate, object);

    } else { // go through all found triples and add reversed one

      for(Enumeration en = res.elements(); en.hasMoreElements();) {
        Statement s = (Statement)en.nextElement();
        Resource inv = (Resource)inverse.get(s.predicate());
        if(inv != null && s.object() instanceof Resource)
          derived.add(nodeFactory.createStatement((Resource)s.object(), inv, s.subject()));
      }
    }
    SetOperations.unite(res, derived);

    return res;
  }

  private void collectInverse(Model res, Resource subject, Resource predicate, RDFNode object) throws ModelException {

    // check inverse relationships
    Resource inv = (Resource)inverse.get(predicate);
    if(inv != null) {
      Model res2 = instances.find((Resource)object, inv, subject);
      for(Enumeration en = res2.elements(); en.hasMoreElements();) {
        Statement s = (Statement)en.nextElement();
        if(s.object() instanceof Resource)
          res.add(nodeFactory.createStatement((Resource)s.object(), predicate, s.subject()));
      }
    }
    //    System.out.println("COLLECTING inverse for " + subject + " " + predicate + " " + object + ": " + res.size());
  }

  /**
   * Search for a single triple.
   * <code>null</code> input for any parameter will match anything.
   *
   * @return  a single triple if such was found;<br>
   *          <code>null</code> if nothing was found;<br>
   *          otherwise a RuntimeException is thrown.
   public Statement find1( Resource subject, Resource predicate, RDFnode object );
  */

  /**
   * Clone the model.
   */
  public Model duplicate() throws ModelException {
    return new UMLModel(nodeFactory, instances.duplicate(), schema, inverse);
  }

  /**
   * Creates empty model of the same Class
   */
  public Model create() throws ModelException {
    return new UMLModel(nodeFactory, instances.create(), schema, inverse);
  }

  /**
   * Returns the node factory for this model
   */
  public NodeFactory getNodeFactory() throws ModelException {
    return nodeFactory;
  }

  public String toString() {

    try {
      return "[UMLModel " + getURI() + "]";
    } catch (ModelException exc) {
      return "[UMLModel: " + exc + "]";
    }
  }




  /**
   * Validates the model given instance and schema data
   */
  public static void validateRawSchema(Model instances, Model schema) throws ModelException {

    //    instances = new OrderedModel(instances);
    //    schema = new OrderedModel(schema);

    StringBuffer errors = new StringBuffer();
    // indicates that cardinality of property has beed verified
    Hashtable verifiedCardinality = new Hashtable();

    UMLModel umlizedSchema = new UMLModel(schema, schema);
    Model umlizedInstance = new UMLModel(instances, schema);
    Model rdfsClosure = RDFSchemaModel.computeRDFSClosure(getRDFSClassHierarchy(umlizedSchema));

    Model rdfSchema = new RDFSchemaModel(schema, rdfsClosure);
    Model rdfInstance = new RDFSchemaModel(instances, rdfsClosure);
    Hashtable inverse = umlizedSchema.getInverse();

    //    RDFUtil.printStatements(umlizedInstance.find(null, DataTypes.Multiplicity_range, null), System.out);

    for(Enumeration en = instances.elements(); en.hasMoreElements();) {
      
      Statement t = (Statement)en.nextElement();
      
      // rdf:type
      
      if(RDF.type.equals(t.predicate())) {
        // ensure that the target is of type core:Classifier
        // ensure that the target is of type rdf:Class
        if(t.object() instanceof Literal)
          invalid( errors, t, "Literals cannot be used for typing" );

	//	if(DataTypes.PseudostateKind.getLabel().equals(t.object().getLabel())) {
	//	  System.err.println("TEST: ");
	//	  RDFUtil.printStatements(rdfSchema.find(DataTypes.PseudostateKind, RDF.type, UMLCore.Classifier), System.out);
	  //    RDFUtil.printStatements(rdfsClosure.find(UMLCore.DataType, null, null), System.out);
	  //	}

        Model res = rdfSchema.find( (Resource)t.object(), RDF.type, UMLCore.Classifier );
        if(res.isEmpty()) {
          //      System.err.println("SUPERCLASSES(C)");
          //      RDFUtil.printStatements( rdfSchema.find( (Resource)t.object(), RDF.type, null ), System.err);
	  invalid( errors, t, t.object().toString() + " must be an instance of " + UMLCore.Classifier);
        }


	// t.subject() is typed. Check cardinality (multiplicity) of incoming and outgoing arcs
	Resource subject = t.subject();
	Resource subjectType = (Resource)t.object();

	if(!verifiedCardinality.contains(subject)) {

	  verifiedCardinality.put(subject, subject);
	  // get association ends that are typed with subjectType
	  Model ends = umlizedSchema.find(null, UMLCore.AssociationEnd_type, subjectType);
	  for(Enumeration endsEn = ends.elements(); endsEn.hasMoreElements();) {
	    Resource end = ((Statement)endsEn.nextElement()).subject();
	    // multiplicity defined? Need to get multiplicity from the opposite AssociationEnd
	    Resource otherEnd = (Resource)umlizedSchema.getInverse().get(end);
	    if(otherEnd == null) // don't complain now, we'll check later for that
	      continue;
	    Resource multiplicity = RDFUtil.getObjectResource(umlizedSchema, otherEnd, UMLCore.AssociationEnd_multiplicity);
	    if(multiplicity != null) {
	      // determine multiplicity of this relation
	      //	      System.out.println("CHECK: " + subject + " (of type " + subjectType + ") for outgoing link " + otherEnd + ": ");
	      Model allEnds = umlizedInstance.find(subject, otherEnd, null);

	      // verify declared multiplicity of this guy: find a range where the actual number of links fits
	      Model ranges = umlizedSchema.find(multiplicity, DataTypes.Multiplicity_range, null);
	      //	      System.out.println("Multiplicity: " + multiplicity + ", " + ranges.size());
	      boolean foundRange = false;

	      String expectedRange = "";
	      for(Enumeration enRanges = ranges.elements(); enRanges.hasMoreElements();) {

		try {
		  Resource range = (Resource)((Statement)enRanges.nextElement()).object();
		  // find lower and upper limit of this range
		  String rangeLower = RDFUtil.getObjectLiteral(umlizedSchema, range, DataTypes.MultiplicityRange_lower);
		  //		  System.out.println("Lower: " + rangeLower);
		  RDFNode rangeUpper = RDFUtil.getObject(umlizedSchema, range, DataTypes.MultiplicityRange_upper);
		  //		  System.out.println("Upper: " + rangeUpper);
		  int lower = Integer.parseInt(rangeLower);
		  int upper = rangeUpper instanceof Resource ? Integer.MAX_VALUE : Integer.parseInt(rangeUpper.getLabel());
		  if(!"".equals(expectedRange))
		    expectedRange += ",";
		  if(lower == upper)
		    expectedRange += rangeLower;
		  else
		    expectedRange += rangeLower + "-" + ((rangeUpper instanceof Resource) ? "*" : rangeUpper.getLabel());
		  if(lower <= allEnds.size() && allEnds.size() <= upper) {
		    foundRange = true;
		    break;
		  }
		} catch (Exception any) {
		  //		  System.err.println("EXC: " + any);
		}
	      }
	      
	      if(!"".equals(expectedRange) && !foundRange) {
		//		System.err.println("--------- ERROR: Resource " +  subject + " (of type " + subjectType + ") is expected to have " + expectedRange + ", not " + allEnds.size() +", properties of type " + otherEnd + " (inverse of " + end + ")");
		invalid( errors, null, "Resource " +  subject + " (of type " + subjectType + ") is expected to have " + expectedRange + ", not " + allEnds.size() +", properties of type " + otherEnd + " (inverse of " + end + ")");
	      }
	    }
	  }
	}

      } else {

        // is t.predicate() an Association?
        Resource rangeType = RDFUtil.getObjectResource(rdfSchema, t.predicate(), UMLCore.AssociationEnd_type);

        if(rangeType != null) {
	  // check domain and range

	  if(!isInstanceOf(rdfInstance, rdfSchema, (Resource)t.object(), rangeType)) {
	    
	    invalid( errors, t, "Object must be instance of " + rangeType );
	  }
	  
	  // find all allowed domains of the Property
	  Resource inverseEnd = (Resource)inverse.get(t.predicate());
	  if(inverseEnd == null)
	    invalid( errors, null, "Bad schema definition of " + t.predicate() + ", no other AssociationEnd found");
	  else {
	    
	    Resource domainType = RDFUtil.getObjectResource(rdfSchema, inverseEnd, UMLCore.AssociationEnd_type);
	    if(domainType == null)
	      invalid( errors, null, "No association end type defined for " + inverseEnd);
	    else if (!isInstanceOf(rdfInstance, rdfSchema, t.subject(), domainType)) {
	      
	      invalid( errors, t, "Subject must be instance of " + domainType );
	    }
	  }
	} else {
	  // is t.predicate() an Attribute?
	  Resource attType = RDFUtil.getObjectResource(rdfSchema, t.predicate(), UMLCore.StructuralFeature_type);

	  if(attType != null) {

	    // try to check for various known Data Types
	    if(!checkDataType(errors, t, attType)) {
	      // check for non data-types

	      if(!(t.object() instanceof Resource) ||
		 !isInstanceOf(rdfInstance, rdfSchema, (Resource)t.object(), attType))
	    
		invalid( errors, t, "Attribute value must be instance of " + attType );
	    }
	  }
	}
	
      }
    }
    if(errors.length() > 0) {
      throw new InvalidModelException(errors.toString());
    }
  }

  static boolean isInstanceOf(Model instance, Model schema, Resource r, Resource umlClass) throws ModelException {

    return
      RDFUtil.isInstanceOf(instance, r, umlClass) ||
      RDFUtil.isInstanceOf(schema, r, umlClass);
  }

  static boolean isValidNumber(String str) {

    try {
      new java.math.BigInteger(str);
      return true;
    } catch (NumberFormatException nfe) {
      return false;
    }
  }

  static boolean checkDataType(StringBuffer errors, Statement t, Resource attType) throws ModelException {

    RDFNode value = t.object();
    String strVal = value.getLabel();

    if (DataTypes.Integer.equals(attType)) {

      if(!(value instanceof Literal) || !isValidNumber(strVal))
	invalid( errors, t, "Attribute value must be an integer.");

    } else if (DataTypes.UnlimitedInteger.equals(attType)) {

      if(!DataTypes.UnlimitedInteger_c42.equals(value) &&
	 (!(value instanceof Literal) || !isValidNumber(strVal)))
	invalid( errors, t, "Attribute value must be an integer or " + DataTypes.UnlimitedInteger_c42);

    } else
      return false;

    return true;
  }

    

  static void invalid(StringBuffer b, Statement t, String msg) {

    if(b.length() > 0)
      b.append('\n');
    if(t != null) {
      //      String objStr = (t.object() instanceof Literal) ? "\"" + t.object() + "\"" : t.object().toString();
      //      b.append("INVALID STATEMENT: (" + t.subject() + ", " + t.predicate() + ", " + objStr + "). ");
      b.append("Invalid statement:\n\t" + t + ".\n\t");
    }
    b.append(msg);
  }

  public static void main(String[] args) throws Exception {

    // testing closure
    if(args.length == 0) {
      System.err.println("UMLModel <instance URL> {<schema URL>}+");
      System.exit(1);
    }

    RDFFactory f = new RDFFactoryImpl();
    Model m = f.createModel();
    Model rawSchema = f.createModel();
    RDFSchemaModel.readModelsFromArgList(f, m, rawSchema, args);

    //    RDFUtil.printStatements(m, System.out);

    try {
      UMLModel.validateRawSchema(m, rawSchema);
      System.err.println("Model is valid.");

    } catch (InvalidModelException ime) {
      System.err.println(ime.getMessage());
    }
  }
        
}
